@@ -24,6 +24,8 @@ Block Read/Write Functions
 
 from contextlib import contextmanager
 
+from .utils.log import FFmpegError
+from .utils.filter import FilterGraph
 from .transcode import transcode
 from . import caps, probe, audio, image, video
 from . import ffmpegprocess as process
@@ -31,7 +33,7 @@ from . import ffmpeg as _ffmpeg
 from . import streams as _streams
 
 __all__ = ["transcode", "caps", "probe", "set_path", "audio", "image", "video", "open", "process"]
-__version__ = "0.0.14"
+__version__ = "0.1.0"
 
 ffmpeg_info = _ffmpeg.versions
 set_path = _ffmpeg.find
@@ -170,7 +172,7 @@ def open(
     # filter = "f" in mode
     # backwards = "b" in mode
     unk = set(mode) - set("avrw")
-    if unk :
+    if unk:
         raise Exception(
             f"Invalid FFmpeg streaming mode: {mode}. Unknown mode {unk} specified."
         )
@@ -235,15 +237,10 @@ def open(
             if audio
             else (_streams.SimpleVideoWriter if write else _streams.SimpleVideoReader)
         )
-        if read:
-            kwds["stream_id"] = stream_id or 0
         if write:
             kwds["rate"] = rate
             kwds["dtype"] = dtype
-            if video:
-                kwds["shape"] = shape
-            elif audio:
-                kwds["channels"] = channels
+            kwds["shape"] = shape or channels
 
     # instantiate the streaming object
     # TODO wrap in try-catch if AV stream fails to try a multi-stream version
